#ifndef WL_WL_HPP_
#define WL_WL_HPP_

/**
 * Graph
 */

#include "wl/details/edge_colored_graph.hpp"

/**
 * A fast implementation of 1-WL
 */

#include "wl/details/canonical_color_refinement.hpp"

/**
 * A alternative implementation of 1-WL and 2-FWL
 */

#include "wl/details/weisfeiler_leman.hpp"
#include "wl/details/weisfeiler_leman_1d.hpp"
#include "wl/details/weisfeiler_leman_2d.hpp"

#endif
